/* 
 * Copyright (C) 2019 Vad Rulezz and developers
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 *
 * This loader is heavily based on flif-pixbuf-loader sources,
 * that list following authors and copyrights:
 *
 * Copyright (C) 2011 Alberto Ruiz
 * Copyright (C) 2011 David Mazary
 * Copyright (C) 2014 P.emysl Janouch
 * Copyright (C) 2017 Leo Izen (thebombzen)
 *
 * Authors: Alberto Ruiz <aruiz@gnome.org>
 *          David Mazary <dmaz@vt.edu>
 *          P.emysl Janouch <p.janouch@gmail.com>
 *          Leo Izen (thebombzen) <leo.izen@gmail.com>
 */

#define GDK_PIXBUF_ENABLE_BACKEND
#include <gdk-pixbuf/gdk-pixbuf.h>
#undef GDK_PIXBUF_ENABLE_BACKEND

#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <libbpg.h>

#ifndef open_memstream
FILE *open_memstream(char **ptr, size_t *sizeloc);
#endif

/* Progressive loader context */
typedef struct {
        GdkPixbufModuleSizeFunc size_func;
        GdkPixbufModuleUpdatedFunc update_func;
        GdkPixbufModulePreparedFunc prepare_func;
 
        guint32 w;
        guint32 h;
        gint bit_depth;
        gint nb_channels;
 
        gpointer user_data;
        GdkPixbuf *pixbuf;
        gboolean got_header;
        GError **error;
 
        FILE *increment_buffer;
	char *increment_buffer_ptr;
        size_t increment_buffer_size;
} BPGContext;


static void destroy_data(guchar *pixels, gpointer data) {
  g_free(pixels);
}

static GdkPixbuf *gdk_pixbuf__bpg_image_load_buf(gpointer data, guint data_size, GError **error) {
  BPGDecoderContext *dec = NULL;
  BPGImageInfo dec_info;
  int w, h, y;
  GdkPixbuf * volatile pixbuf = NULL;
  gpointer decdata;
  gint ok;
  uint32_t *dst;

  dec = bpg_decoder_open();

  if (bpg_decoder_decode(dec, (uint8_t*)data, data_size) < 0) {
	g_set_error(error, GDK_PIXBUF_ERROR, GDK_PIXBUF_ERROR_FAILED,
                "Failed to decode image");
	bpg_decoder_close( dec );
        return FALSE;
  }

  bpg_decoder_get_info( dec, &dec_info );
  w = dec_info.width;
  h = dec_info.height;

  decdata = g_malloc( w * h * 4 );

  bpg_decoder_start( dec, BPG_OUTPUT_FORMAT_RGBA32 );

  dst = (uint32_t*)decdata;
  for ( y = 0; y < h; y++ ) {
	bpg_decoder_get_line( dec, dst );
	dst += w;
  }

  bpg_decoder_close( dec );

  pixbuf = gdk_pixbuf_new_from_data (
	(const guchar *)decdata, GDK_COLORSPACE_RGB, /* has_alpha */ TRUE,
        8, w, h, w * 4, destroy_data, NULL);

  if (!pixbuf) {
        g_set_error (error, GDK_PIXBUF_ERROR, GDK_PIXBUF_ERROR_FAILED,
                "Failed to decode image");
        return FALSE;
  }

  return pixbuf;
}

static GdkPixbuf *gdk_pixbuf__bpg_image_load(FILE *f, GError **error) {
  guint data_size;
  gpointer data;
  gboolean ok;

  fseek (f, 0, SEEK_END);
  data_size = ftell(f);
  fseek (f, 0, SEEK_SET);

  /* Get data */
  data = g_malloc(data_size);
  ok = (fread(data, data_size, 1, f) == 1);
  if (!ok) {
        g_free( data );
	g_set_error(error, GDK_PIXBUF_ERROR, GDK_PIXBUF_ERROR_FAILED,
	        "Failed to read file");
        return FALSE;
  }
  fclose( f );

  GdkPixbuf *pixbuf = gdk_pixbuf__bpg_image_load_buf( data, data_size, error );
  g_free( data );

  return pixbuf;
}

static gpointer gdk_pixbuf__bpg_image_begin_load (GdkPixbufModuleSizeFunc size_func, GdkPixbufModulePreparedFunc prepare_func,
                                   GdkPixbufModuleUpdatedFunc update_func, gpointer user_data, GError **error) {

    BPGContext *context = g_new(BPGContext, 1);
    context->size_func = size_func;
    context->prepare_func = prepare_func;
    context->update_func  = update_func;
    context->user_data = user_data;
    context->error = error;
    context->got_header = FALSE;
            
    context->increment_buffer = open_memstream(&context->increment_buffer_ptr, &context->increment_buffer_size);
    if (!context->increment_buffer){
        perror("Cannot create increment buffer.");
        g_free(context);
        return NULL;
    }
    
    return context;
}

static gboolean gdk_pixbuf__bpg_image_stop_load (gpointer user_context, GError **error) {
    BPGContext *context = (BPGContext *) user_context;
 
    int status = fclose(context->increment_buffer);
    if (status != 0){
        perror("Cannot flush/close increment buffer.");
        // we don't use g_free here because open_memstream uses the internal malloc from stdlib.h
        free( context->increment_buffer_ptr );
        g_free(context);
        return FALSE;
    }

    context->pixbuf = gdk_pixbuf__bpg_image_load_buf(context->increment_buffer_ptr, context->increment_buffer_size, error);
    free( context->increment_buffer_ptr );

    if ( !context->pixbuf ) {
    	g_free(context);
	return FALSE;
    }

    if (context->prepare_func)
       	(* context->prepare_func) (context->pixbuf, NULL, context->user_data);
 
    if (context->update_func)
	(* context->update_func) (context->pixbuf, 0, 0, context->w, context->h, context->user_data);

    g_object_unref(context->pixbuf);
    g_free(context);

    return TRUE;
}

static gboolean gdk_pixbuf__bpg_image_load_increment (gpointer user_context, const guchar *buf, guint size, GError **error) {
    BPGContext *context = (BPGContext *) user_context;

    int status = fwrite(buf, size, sizeof(guchar), context->increment_buffer);

    if (status != sizeof(guchar)){
        g_set_error(error, GDK_PIXBUF_ERROR, GDK_PIXBUF_ERROR_FAILED,
            "Can't write to increment buffer.");
        return FALSE;
    }
    
    if (!context->got_header) {
        BPGImageInfo bpg_info;
  	BPGDecoderContext *dec = bpg_decoder_open();

	if (bpg_decoder_decode(dec, (uint8_t*)buf, size) < 0) {
		g_set_error(error, GDK_PIXBUF_ERROR, GDK_PIXBUF_ERROR_FAILED,
	                "Failed to decode image");
		bpg_decoder_close( dec );
	        return FALSE;
	}

	if ( bpg_decoder_get_info(dec, &bpg_info) < 0 ) {
		bpg_decoder_close( dec );
                g_set_error (error, GDK_PIXBUF_ERROR, GDK_PIXBUF_ERROR_CORRUPT_IMAGE,
                    "Cannot read BPG image header.");
                return FALSE;
        }
        
        context->w = bpg_info.width;
        context->h = bpg_info.height;
        context->nb_channels = 3 + (bpg_info.has_alpha? 1: 0);
        context->bit_depth = bpg_info.bit_depth;
        bpg_decoder_close( dec );

        context->got_header = TRUE;

        if (context->size_func) {
            gint scaled_w = context->w;
            gint scaled_h = context->h;

            (* context->size_func) (&scaled_w, &scaled_h, context->user_data);
            if ((guint)scaled_w != context->w || (guint)scaled_h != context->h) {
                context->w = scaled_w;
                context->h = scaled_h;
            }
        }
        
    }   
    return TRUE;
}

int bpg_encode_mem( unsigned char *pixels, int w, int h, int stride, int has_alpha, FILE *f );

static gboolean save_to( GdkPixbufSaveFunc save_func, gpointer user_data, GdkPixbuf *pixbuf,
        gchar **keys, gchar **values, GError **error ) {
	gboolean result;
	char *buf = 0;
       	size_t len;

	FILE *f = open_memstream( &buf, &len );

	result = bpg_encode_mem( gdk_pixbuf_get_pixels( pixbuf ),
		gdk_pixbuf_get_width( pixbuf ), gdk_pixbuf_get_height( pixbuf ),
	        gdk_pixbuf_get_rowstride( pixbuf ), gdk_pixbuf_get_has_alpha( pixbuf ), f );

        if ( result )
                save_func( buf, len, 0, user_data );
        else
                g_set_error( error, GDK_PIXBUF_ERROR, GDK_PIXBUF_ERROR_FAILED,
			"Cannot encode picture as BPG" );

        free( buf );

	return result;
}

void fill_vtable(GdkPixbufModule *module) {
    module->load = gdk_pixbuf__bpg_image_load;
    module->begin_load = gdk_pixbuf__bpg_image_begin_load;
    module->stop_load = gdk_pixbuf__bpg_image_stop_load;
    module->load_increment = gdk_pixbuf__bpg_image_load_increment;
    module->save_to_callback = save_to;
}

void fill_info(GdkPixbufFormat *info) {
        static GdkPixbufModulePattern signature[] = {
                { "BPG", "   ", 100 },
                { NULL, NULL, 0 }
        };

        static gchar *mime_types[] = {
                "image/bpg", NULL
        };

        static gchar *extensions[] = {
                "bpg", "BPG", NULL
        };

        info->name        = "bpg";
        info->signature   = signature;
        info->description = "The BPG image format";
        info->mime_types  = mime_types;
        info->extensions  = extensions;
        info->flags       = GDK_PIXBUF_FORMAT_WRITABLE | GDK_PIXBUF_FORMAT_THREADSAFE;
        info->license     = "LGPL";
}
